import os
import sys
import threading

from ..util import cw


class Spinner:
    def __init__(
        self,
        message: str | None = None,
        color: bool = False,
        sequence: list[str] | str = "⠋⠙⠚⠞⠖⠦⠴⠲⠳⠓",
        stream=sys.stdout,
        final: str = "✓",
        final_fail: str = "✗",
        delay: float = 0.25,
        newline: bool | None = None
    ):
        self._message = message
        self._newline = newline if newline is not None else message is not None
        self._sequence = sequence if isinstance(sequence, list) else list(sequence)
        self._stream = stream
        self._final_ok = final
        self._final_fail = final_fail
        self._delay = delay

        self._result: bool | None = None
        self._thread: threading.Thread | None = None
        self._stopped = False

        self._c = cw(color)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stop()

    def start(self):
        def spin():
            i = 0

            while not self.stop_running.is_set():
                self._stream.write(self._c(self._sequence[i], "yellow"))
                self._stream.flush()
                self.stop_running.wait(self._delay)
                n = len(self._sequence[i])
                self._stream.write("\b" * n + " " * n + "\b" * n)
                self._stream.flush()

                i = i + 1
                if len(self._sequence) <= i:
                    i = 0

        if self._message is not None:
            self._stream.write(self._message)

        if os.isatty(self._stream.fileno()):
            self.stop_running = threading.Event()
            self._thread = threading.Thread(target=spin)
            self._thread.start()

    def _stop(self, final: str | None = None):
        if final is None:
            final = (
                self._c(self._final_ok, "green")
                if self._result
                else self._c(self._final_fail, "red")
            )

        if not self._stopped:
            self._stopped = True

            if self._thread is not None:
                self.stop_running.set()
                self._thread.join()

            if self._result is not None:
                self._stream.write(final)
                self._stream.flush()

            if self._newline:
                self._stream.write("\n")
                self._stream.flush()

    def ok(self, result=True):
        self._result = result

    def fail(self):
        self._result = False

    def stop(self, result: bool | str | None = None):
        if isinstance(result, str):
            self._stop(result)
        elif result is not None:
            self._result = result
        self._stop()



import os.path as path
import sys
from dataclasses import dataclass
from pathlib import Path


@dataclass
class Config:
    cpp_libs_dir: str | None = None
    extra_flags: dict[str, list[str]] | None = None
    color: bool = sys.stdout.isatty()
    diff: bool = True
    ignore_debug: bool = True
    verbose: bool = False


def find_config(dir: str) -> str | None:
    p = Path(path.abspath(dir))

    while not path.exists(path.join(p, ".nkconfig.py")):
        if p.parent == p:
            return None

        p = p.parent

    return path.join(p, ".nkconfig.py")


def exec_config(dir: str) -> Config:
    cfg_path = find_config(dir)

    cfg = Config()

    if cfg_path is not None:
        with open(cfg_path, "r") as script:
            exec(script.read(), {"cfg": cfg, "__file__": cfg_path})

    return cfg



import os
import os.path as path
import re
import shutil
import tempfile
import zipfile
from dataclasses import dataclass
from typing import Literal

import appdirs
import natsort
import requests

from .config import Config
from .interactive import spinner


@dataclass
class ProblemIO:
    """
    Stores the paths to input, output and a source name.
    """

    i: str
    o: str | None
    from_: Literal["sample"] | Literal["local"]

    def input_name(self):
        return path.basename(self.i)


def problem_dir(problem: str) -> str:
    return path.join(appdirs.user_cache_dir("nekontrol"), "problems", problem)


def download_tests(problem: str, config: Config) -> bool:
    url = f"https://open.kattis.com/problems/{problem}/file/statement/samples.zip"

    problem_cache_dir = problem_dir(problem)

    if path.exists(problem_cache_dir):
        print(f"Using cached sample data for problem {problem}")
        return True
    else:
        with spinner.Spinner(
            f"Downloading sample data for problem {problem} ", color=config.color
        ) as s:
            with requests.get(url, stream=True) as response:
                if not response.ok:
                    s.fail()
                    return False
                with tempfile.TemporaryFile("w+b") as f:
                    shutil.copyfileobj(response.raw, f)
                    f.seek(0)
                    zip = zipfile.ZipFile(f)
                    zip.extractall(problem_cache_dir)
                    s.ok()
                    return True


def problem_sample_inputs_outputs(
    problem: str, config: Config
) -> list[ProblemIO] | None:
    problem_cache_dir = problem_dir(problem)
    downloaded = download_tests(problem, config)

    if not downloaded:
        return None

    in_files = [
        path.join(problem_cache_dir, f)
        for f in os.listdir(problem_cache_dir)
        if f.endswith(".in")
    ]

    def get_out_file(in_file: str) -> str | None:
        out_file = re.sub(r"\.in$", ".ans", in_file)

        return out_file if path.exists(out_file) else None

    ins_and_outs = [(in_file, get_out_file(in_file), "sample") for in_file in in_files]

    return sorted_problems([ProblemIO(*iof) for iof in ins_and_outs])


def local_inputs_outputs(dir: str, name: str) -> list[ProblemIO]:
    file_names = [fn for fn in os.listdir(dir or ".")]

    candidates = [
        (fn, m)
        for fn in file_names
        if (m := re.match("^" + re.escape(name) + r"((?:\.\d+)?)\.in", fn))
    ]

    ins_and_outs: list[ProblemIO] = [
        ProblemIO(
            path.join(dir, in_),
            out
            if path.isfile(out := path.join(dir, name + m.group(1) + ".ans"))
            else None,
            "local",
        )
        for (in_, m) in candidates
    ]

    return sorted_problems(ins_and_outs)


def sorted_problems(
    problems: list[ProblemIO],
) -> list[ProblemIO]:
    return natsort.natsorted(problems, key=lambda p: p.i)



import time

from .. import compare, util
from ..config import Config
from ..language import Language
from ..problems import ProblemIO
from . import spinner


def run(name: str, lang: Language, io: ProblemIO, config: Config):
    c = util.cw(config.color)

    with spinner.Spinner(
        f"Running {name} with {io.from_} input {io.input_name()} ",
        color=config.color,
        newline=False,
    ) as s:
        start = time.time()
        exit_code, stdout, stderr = lang.run(io.i)
        finish = time.time()
        duration = finish - start

        if duration < 1:
            time_color_args = dict(color="black", on_color="on_green")
        elif duration < 3:
            time_color_args = dict(color="black", on_color="on_yellow")
        else:
            time_color_args = dict(color="black", on_color="on_red")
        time_msg = " " + c(f" ⏱  {duration:.3} s", **time_color_args)

        diff = None

        if config.diff and io.o is not None:
            with open(io.o, "r") as expected:
                diff = compare.compare_outputs(stdout, expected.read(), io.i, config)

                s.stop(isinstance(diff, bool))
                print(time_msg)

                if isinstance(diff, str):
                    print(diff)
                elif diff is True:
                    print(c("NOTE: The output contained debug lines", "yellow"))

                if exit_code != 0:
                    print(
                        c(
                            f"Proccess exited with a non-zero exit code {exit_code}"
                            + (" and the following stderr:" if stderr else ""),
                            "red",
                        )
                    )

                    if stderr:
                        print(util.indented(stderr))
        else:
            s.stop()
            print(time_msg)

            print(c("Input:", "yellow"))
            with open(io.i, "r") as ifile:
                print(util.indented(ifile.read()))
            print(c("Got output:", "yellow"))
            print(util.indented(stdout))



import difflib
import re

import termcolor

from . import util
from .config import Config

debug_regex = re.compile("^(?:debug|dbg)", re.IGNORECASE)
inline_debug_regex = re.compile(r"\((?:dbg|debug)[^)]*\)", re.IGNORECASE)


def compare_outputs(
        output: str, expected_output: str, input_file: str, config: Config
) -> str | bool:
    c = util.cw(config.color)

    found_debug = False

    if config.ignore_debug:
        output_lines = []
        for line in output.splitlines():
            if debug_regex.match(line):
                found_debug = True
                continue

            inline_line = inline_debug_regex.sub("", line)

            if len(inline_line) != len(line):
                found_debug = True
            line = inline_line

            output_lines.append(line)
    else:
        output_lines = output.splitlines()

    differ = difflib.Differ()
    diff = list(
        differ.compare(
            list(map(str.rstrip, expected_output.splitlines())),
            list(map(str.rstrip, output_lines)),
        )
    )

    if all(d.startswith("  ") for d in diff):
        return found_debug
    else:
        with open(input_file, "r") as ifile:
            return (
                c("Input:", "yellow")
                + "\n"
                + util.indented(ifile.read())
                + "\n"
                + c("Diff:", "yellow")
                + "\n"
                + "\n".join(color_diff(diff) if config.color else diff)
            )


def color_diff(diff: list[str]) -> list[str]:
    def color_line(line: str) -> str:
        match line[0:2]:
            case "+ " | "- " as start:
                return termcolor.colored(start, "red") + line[2:]
            case "? " as start:
                return termcolor.colored(start, "yellow") + line[2:]
            case _:
                return line

    return list(map(color_line, diff))



import os.path as path

import click

from .. import language, problems, util
from ..config import exec_config
from . import run

executable_file = click.Path(exists=True, readable=True, file_okay=True, dir_okay=False)


@click.command(context_settings={"help_option_names": ["-h", "--help"]})
@click.argument("file-path", metavar="FILE", type=executable_file)
@click.option("-p", "--problem", type=str, help="The kattis problem name")
@click.option("-d", "--diff", type=bool)
@click.option("-c", "--color", type=bool)
@click.option("-v", "--verbose", type=bool)
@click.option("--ignore-debug", type=bool)
def cli(
    file_path: str,
    problem: str | None,
    color: bool | None,
    diff: bool | None,
    ignore_debug: bool | None,
    verbose: bool | None,
):
    """nekontrol - Control your kattis solutions.

    Run FILE and test against sample and local test data.
    """
    file_name = path.basename(file_path)
    file_dir = path.dirname(file_path)
    file_base, extension = path.splitext(file_name)

    config = exec_config(file_dir)
    if color is not None:
        config.color = color
    if diff is not None:
        config.diff = diff
    if ignore_debug is not None:
        config.ignore_debug = ignore_debug
    if verbose is not None:
        config.verbose = verbose

    c = util.cw(config.color)

    if problem is None:
        if config.verbose:
            print(c(f"No problem name specified, guessing '{file_base}'", "yellow"))
        problem = file_base

    local_ios = problems.local_inputs_outputs(file_dir, file_base)
    sample_ios = problems.problem_sample_inputs_outputs(problem, config)
    ios = local_ios + (sample_ios or [])

    if len(ios) == 0:
        raise click.ClickException(f"Found no inputs to run for problem {problem}")

    lang = language.get_lang(file_path, config)

    if lang is None:
        raise click.ClickException(
            f"Language for file extension {extension} is not implemented."
        )

    with lang:
        for io in ios:
            run.run(file_name, lang, io, config)



import os
import tempfile
from . import config

import termcolor


class TempFileName:
    def __init__(self, *args):
        self.args = args

    def __enter__(self) -> str:
        fd, file_name = tempfile.mkstemp(
            ".exe" if os.name == "nt" else None, *self.args
        )
        os.close(fd)
        self.file_name = file_name
        return self.file_name

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.remove(self.file_name)


def indented(s: str) -> str:
    return "\n".join("  " + ds for ds in s.splitlines())


def cw(color: bool):
    """Create a termcolor.colored wrapper that will or will not color the string."""

    def inner(s, *args, **kwargs):
        if color:
            return termcolor.colored(s, *args, **kwargs)
        else:
            return s

    return inner



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['nekontrol', 'nekontrol.interactive']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'click>=8.1.3,<9.0.0',
 'natsort>=8.2.0,<9.0.0',
 'requests>=2.28.2,<3.0.0',
 'termcolor>=2.2.0,<3.0.0']

entry_points = \
{'console_scripts': ['nk = nekontrol.interactive.cli:cli']}

setup_kwargs = {
    'name': 'nekontrol',
    'version': '0.2.2',
    'description': 'A program for testing kattis solutions with input and output.',
    'long_description': '# nekontrol - Control your kattis solutions\n\nThis is a simple program that compiles, runs and tests your kattis solutions\nagainst local and sample input and output.\n\n## Demo\n\n![Demo](https://raw.githubusercontent.com/Quaqqer/nekontrol/main/res/demo.svg)\n\n## Features\n\n- Automatically downloads sample input and output\n- Discovers local sample input and output\n- Compiles your source code and runs it with local and sample input and output\n- Ignores debug messages - ignores lines starting with "dbg" or "debug" and\n  inline messages "(dbg...)" and "(debug...)". If debug lines are discovered\n  then you are notified in the output so that you don\'t submit something that is\n  incorrect.\n\nThe current supported languages are:\n\n- C++ (requires `c++` in your path)\n- Rust (requirest `rustc` in your path)\n- Python (requires python in your path)\n- Haskell (requires `ghc` in your path)\n- Lua (requires `lua` in your path)\n- JavaScript (Node) (requires `node` in your path)\n\n> **Note**\n> Nekontrol will try to find a way to run your program, but it is not guranteed\n> that the program it uses matches the version that kattis has. For instance\n> kattis runs python with pypy 3.8. Nekontrol will try to run it with pypy 3.8,\n> but will fall back to other versions if it cannot be found.\n\nAdding more languages is left as an exercise for the reader.\n\n## Usage\n\nSimply run `nk <source file>` and it should hopefully compile (if needed)\nand run!\n\n- `nk` assumes that the name of the source files matches the name of the\n  problem (found in the url, ex. https://open.kattis.com/problems/<b>hello</b>).\n  If this assumption is incorrect, you can specify the problem with\n  `--problem <problem name>`. Local files are still matched with the filename\n  regardless of the problem specified.\n- Input and output files should follow the format `<filename>.in` or\n  `<filename>.<number>.in` and corresponding outputs are named `<filename>.ans`\n  etc. where `<filename>` comes from `nk <filename>.cpp` for instance.\n\n> **Note**\n> Multiple files are not supported as of yet\n\n## Requirements\n\n- The requirements of this program is only having python 3.10 or newer as well\n  pip\n- A half decent terminal, almost anything other than cmd.exe on windows will\n  work\n\n## Install or update\n\nTo install or update, run the following.\n\n```sh\npip install -U nekontrol\n```\n\nNow it should hopefully work, enjoy!\n\n### Nix\n\nThere is a flake for Nix users, but if you use Nix I trust you know how to use\nflakes.\n\n## Other\n\n- If you are thinking "nekontrol is a kinda weeby name", I ask you: have you\n  ever tried coming up with an original name?\n',
    'author': 'Emanuel S',
    'author_email': 'emanuel@empa.xyz',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Quaqqer/nekontrol',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import os
import shutil
import subprocess
import tempfile
from os import path

from click import ClickException

from . import util
from .config import Config
from .interactive.spinner import Spinner


class Language:
    def __init__(self, source_file: str, config: Config):
        self.source_file = source_file
        self.config = config

    def run(self, input_file: str) -> tuple[int, str, str]:
        raise NotImplementedError()

    def prepare(self):
        pass

    def cleanup(self):
        pass

    def __enter__(self):
        self.prepare()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()


class InterpretedLanguage(Language):
    @property
    def bins(self) -> list[str]:
        raise NotImplementedError()

    def __init__(self, source_file: str, config: Config):
        super().__init__(source_file, config)
        bin = find_bin(self.bins)

        if bin is None:
            raise Exception(f"Binary for language not found, needs one of {self.bins}")

        self.bin = bin

    def run(self, input_file: str):
        return generic_run([self.bin, self.source_file], input_file)


class Python(InterpretedLanguage):
    @property
    def bins(self):
        return [
            "pypy38",
            "pypy3.8",
            "pypy3",
            "python38",
            "python3.8",
            "python3",
            "python",
        ]


class Lua(InterpretedLanguage):
    @property
    def bins(self):
        return ["lua", "luajit"]


class JSNode(InterpretedLanguage):
    @property
    def bins(self):
        return ["node"]


class CompiledLanguage(Language):
    @property
    def cmdline(self) -> list[str]:
        raise NotImplementedError()

    def prepare(self):
        with Spinner(f"Compiling {self.source_file} ", color=self.config.color) as s:
            self.compiled_output = tempfile.mktemp()
            exit_code, stderr = self.compile()

            if exit_code == 0:
                s.stop(True)
            else:
                s.stop(False)
                err_msg = f"Compilation exited with code {exit_code}" + (
                    " and stderr:\n" + util.indented(stderr) if stderr else ""
                )
                raise ClickException(err_msg)

    def cleanup(self):
        if path.exists(self.compiled_output):
            os.remove(self.compiled_output)

    def run(self, input_file: str):
        return generic_run([self.compiled_output], input_file)

    def compile(self) -> tuple[int, str]:
        """Compile the file.

        Returns:
            Nothing if successful, the exit code and the stderr if the exit
            code was non-0.
        """
        p = subprocess.Popen(
            self.cmdline, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE
        )
        _, stderr = p.communicate()
        exit_code = p.returncode
        return exit_code, stderr.decode("utf-8")


class Cpp(CompiledLanguage):
    @property
    def cmdline(self) -> list[str]:
        cmdline = [
            "c++",
            "--std=c++17",
            self.source_file,
            "-o",
            self.compiled_output,
            f"-fdiagnostics-color={'always' if self.config.color else 'never'}",
        ]
        if self.config.cpp_libs_dir is not None:
            cmdline += [f"-I{self.config.cpp_libs_dir}"]
        return cmdline


class Rust(CompiledLanguage):
    @property
    def cmdline(self):
        return [
            "rustc",
            "-O",
            "--crate-type",
            "bin",
            "--edition=2018",
            self.source_file,
            "--color",
            "always" if self.config.color else "never",
            "-o",
            self.compiled_output,
        ]


class Haskell(CompiledLanguage):
    def prepare(self):
        self.temp_out_dir = tempfile.mkdtemp()
        super().prepare()

    def cleanup(self):
        super().cleanup()
        shutil.rmtree(self.temp_out_dir)

    @property
    def cmdline(self):
        return [
            "ghc",
            "-outputdir",
            self.temp_out_dir,
            self.source_file,
            "-o",
            self.compiled_output,
        ]


def get_lang(source_file: str, config: Config) -> Language | None:
    _, ext = path.splitext(source_file)
    match ext:
        case ".cc" | ".cpp" | ".cxx":
            return Cpp(source_file, config)
        case ".py":
            return Python(source_file, config)
        case ".hs":
            return Haskell(source_file, config)
        case ".rs":
            return Rust(source_file, config)
        case ".lua":
            return Lua(source_file, config)
        case ".js":
            return JSNode(source_file, config)
        case _:
            return None


def generic_run(cmdline: list[str], input_file: str) -> tuple[int, str, str]:
    with open(input_file, "r") as input:
        p = subprocess.Popen(
            cmdline,
            stdin=input,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        byte_streams = p.communicate()
        exit_code = p.returncode
        stdout, stderr = [s.decode("utf-8") for s in byte_streams]
        return exit_code, stdout, stderr


def find_bin(options: list[str]) -> str | None:
    for option in options:
        path = shutil.which(option)
        if path is not None:
            return path
    return None


def compile_generic(cmdline: list[str]) -> None | str:
    p = subprocess.Popen(cmdline, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE)
    assert p.stderr
    _, errbytes = p.communicate()
    return_code = p.wait()

    if return_code == 0:
        return
    else:
        return errbytes.decode("utf-8")
